#include <Wire.h>
#include <SparkFun_WM8960_Arduino_Library.h> 
#include <algorithm>
#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <iostream>
#include <string>

WM8960 codec;
  

// Include I2S driver
#include <driver/i2s.h>

// Connections to I2S
#define I2S_WS 25
#define I2S_SD 17
#define I2S_SDO 4
#define I2S_SCK 16

// Use I2S Processor 0
#define I2S_PORT I2S_NUM_0

// Define input buffer length
#define bufferLen 512
int16_t sBuffer[bufferLen];

// Define Sample Rate
#define SAMPLERATE 44100

// WiFi login credentials
const char* ssid = "SDNet";
const char* password = "CapstoneProject";

// Variables to read the slider values from the web interface
String distortionValue = "0";
String tremoloValue = "0";
String fuzzValue = "0";
String eightBitValue = "0";

const char* PARAM_INPUT = "value";

// Create AsyncWebServer object on port 80
AsyncWebServer server(80);

// Create the visual display of the webpage
const char index_html[] PROGMEM = R"rawliteral(

<!DOCTYPE html>

<head>
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <title>Guitar Effects</title>
  <style>
    html {font-family: Arial; display: inline-block; text-align: center;}
    h2 {font-size: 2.3rem;}
    p {font-size: 1.9rem;}
    body {max-width: 400px; margin:0px auto; padding-bottom: 25px;}
    .slider { -webkit-appearance: none; margin: 14px; width: 360px; height: 25px; background: #2196F3;
      outline: none; -webkit-transition: .2s; transition: opacity .2s;}
    .slider::-webkit-slider-thumb {-webkit-appearance: none; appearance: none; width: 35px; height: 35px; background: #003249; cursor: pointer;}
    .slider::-moz-range-thumb { width: 35px; height: 35px; background: #003249; cursor: pointer; } 
    .slider-container { display: flex; align-items: center; };
    .slider-value { display: inline; margin-left: 10px; };
  </style>
</head>

<body>

  <h2>Distortion</h2>
  <div class="slider-container">
  <label for="distortion-slider"></label>
  <p><input type="range" onchange="updateDistortionSlider(this)" id="distortion-slider" min="0" max="10" value="%DISTORTIONVALUE%" step="1" class="slider"></p>
  <p><span id="textDistortionValue">%DISTORTIONVALUE%</span></p>
  </div>
  
  <h2>Tremolo</h2>
  <div class="slider-container">
  <label for="tremolo-slider"></label>
  <p><input type="range" onchange="updateTremoloSlider(this)" id="tremolo-slider" min="0" max="10" value="%TREMOLOVALUE%" step="0.5" class="slider"></p>
  <p><span id="textTremoloValue">%TREMOLOVALUE%</span></p>
  </div>
  
  <h2>Fuzz</h2>
  <div class="slider-container">
  <label for="fuzz-slider"></label>
  <p><input type="range" onchange="updateFuzzSlider(this)" id="fuzz-slider" min="0" max="10" value="%FUZZVALUE%" step="1" class="slider"></p>
  <p><span id="textFuzzValue">%FUZZVALUE%</span></p>
  </div>

  <h2>8-Bit</h2>
  <div class="slider-container">
  <label for="eightBit-slider"></label>
  <p><input type="range" onchange="updateEightBitSlider(this)" id="eightBit-slider" min="0" max="1" value="%EIGHTBITVALUE%" step="1" class="slider"></p>
  <p><span id="textEightBitValue">%EIGHTBITVALUE%</span></p>
  </div>


  <script>

function updateDistortionSlider(element) {
      var distortionValue = document.getElementById("distortion-slider").value;
      var fuzzValue = document.getElementById("fuzz-slider").value;
      var eightBitValue = document.getElementById("eightBit-slider").value;
      if (fuzzValue != 0) { 
        document.getElementById("fuzz-slider").value = 0;
        document.getElementById("textFuzzValue").innerHTML = 0;
        fuzzValue = 0;
        console.log(fuzzValue);
        var xhr = new XMLHttpRequest();
        xhr.open("GET", "/fuzz-slider?value=" + fuzzValue, true);
        xhr.send();
      }
      if (eightBitValue != 0) { 
        document.getElementById("eightBit-slider").value = 0;
        document.getElementById("textEightBitValue").innerHTML = 0;
        eightBitValue = 0;
        console.log(eightBitValue);
        var xhr = new XMLHttpRequest();
        xhr.open("GET", "/eightBit-slider?value=" + eightBitValue, true);
        xhr.send();
      }
      document.getElementById("textDistortionValue").innerHTML = distortionValue;
      console.log(distortionValue);
      var xhr = new XMLHttpRequest();
      xhr.open("GET", "/distortion-slider?value=" + distortionValue, true);
      xhr.send();
    }
    
    function updateTremoloSlider(element) {
      var tremoloValue = document.getElementById("tremolo-slider").value;
      var eightBitValue = document.getElementById("eightBit-slider").value;
      if (eightBitValue != 0) { 
        document.getElementById("eightBit-slider").value = 0;
        document.getElementById("textEightBitValue").innerHTML = 0;
        eightBitValue = 0;
        console.log(eightBitValue);
        var xhr = new XMLHttpRequest();
        xhr.open("GET", "/eightBit-slider?value=" + eightBitValue, true);
        xhr.send();
      }
      document.getElementById("textTremoloValue").innerHTML = tremoloValue;
      console.log(tremoloValue);
      var xhr = new XMLHttpRequest();
      xhr.open("GET", "/tremolo-slider?value=" + tremoloValue, true);
      xhr.send();
    }

    function updateFuzzSlider(element) {
      var fuzzValue = document.getElementById("fuzz-slider").value;
      var distortionValue = document.getElementById("distortion-slider").value;
      var eightBitValue = document.getElementById("eightBit-slider").value;
      if (distortionValue != 0) { 
        document.getElementById("distortion-slider").value = 0;
        document.getElementById("textDistortionValue").innerHTML = 0;
        distortionValue = 0;
        console.log(distortionValue);
        var xhr = new XMLHttpRequest();
        xhr.open("GET", "/distortion-slider?value=" + distortionValue, true);
        xhr.send();
      }
      if (eightBitValue != 0) { 
        document.getElementById("eightBit-slider").value = 0;
        document.getElementById("textEightBitValue").innerHTML = 0;
        eightBitValue = 0;
        console.log(eightBitValue);
        var xhr = new XMLHttpRequest();
        xhr.open("GET", "/eightBit-slider?value=" + eightBitValue, true);
        xhr.send();
      }
      document.getElementById("textFuzzValue").innerHTML = fuzzValue;
      console.log(fuzzValue);
      var xhr = new XMLHttpRequest();
      xhr.open("GET", "/fuzz-slider?value=" + fuzzValue, true);
      xhr.send();
    }

    function updateEightBitSlider(element) {
      var eightBitValue = document.getElementById("eightBit-slider").value;
      var distortionValue = document.getElementById("distortion-slider").value;
      var tremoloValue = document.getElementById("tremolo-slider").value;
      var fuzzValue = document.getElementById("fuzz-slider").value;
      if (distortionValue != 0) { 
        document.getElementById("distortion-slider").value = 0;
        document.getElementById("textDistortionValue").innerHTML = 0;
        distortionValue = 0;
        console.log(distortionValue);
        var xhr = new XMLHttpRequest();
        xhr.open("GET", "/distortion-slider?value=" + distortionValue, true);
        xhr.send();
      }
      if (tremoloValue != 0) {
        document.getElementById("tremolo-slider").value = 0;
        document.getElementById("textTremoloValue").innerHTML = 0;
        tremoloValue = 0;
        console.log(tremoloValue);
        var xhr = new XMLHttpRequest();
        xhr.open("GET", "/tremolo-slider?value=" + tremoloValue, true);
        xhr.send();
      }
      if (fuzzValue != 0) { 
        document.getElementById("fuzz-slider").value = 0;
        document.getElementById("textFuzzValue").innerHTML = 0;
        fuzzValue = 0;

        console.log(fuzzValue);
        var xhr = new XMLHttpRequest();
        xhr.open("GET", "/fuzz-slider?value=" + fuzzValue, true);
        xhr.send();
      }
      document.getElementById("textEightBitValue").innerHTML = eightBitValue;
      console.log(eightBitValue);
      var xhr = new XMLHttpRequest();
      xhr.open("GET", "/eightBit-slider?value=" + eightBitValue, true);
      xhr.send();
    }

  </script>

</body>

</html>

)rawliteral";


// Replaces placeholder with button section in your web page
String processor(const String& var){

  if (var == "DISTORTIONVALUE"){
    return distortionValue;
  }
  
  if (var == "TREMOLOVALUE"){
    return tremoloValue;
  }

  if (var == "FUZZVALUE"){
    return fuzzValue;
  }

  if (var == "EIGHTBITVALUE"){
    return eightBitValue;
 }

  return String();  
} // ends String processor()

// All the code for the web interface setup and functionality, called in setup()
void web_interface_setup()
{
  // Connect to Wi-Fi
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Connecting to WiFi..");
  }

  // Print ESP Local IP Address
  Serial.println(WiFi.localIP());

  // Route for root / web page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/html", index_html, processor);
  });

  server.on("/distortion-slider", HTTP_GET, [] (AsyncWebServerRequest *request) {
    String inputMessage;
    // GET input1 value on <ESP_IP>/slider?value=<inputMessage>
    if (request->hasParam(PARAM_INPUT)) {
      inputMessage = request->getParam(PARAM_INPUT)->value();
      distortionValue = inputMessage;
    }
    else {
      inputMessage = "No message sent";
    }
    Serial.println("Distortion Value: " + inputMessage);
    request->send(200, "text/plain", "OK");
  });

  // Send a GET request to <ESP_IP>/slider?value=<inputMessage>
  server.on("/tremolo-slider", HTTP_GET, [] (AsyncWebServerRequest *request) {
    String inputMessage;
    // GET input1 value on <ESP_IP>/slider?value=<inputMessage>
    if (request->hasParam(PARAM_INPUT)) {
      inputMessage = request->getParam(PARAM_INPUT)->value();
      tremoloValue = inputMessage;
    }
    else {
      inputMessage = "No message sent";
    }
    Serial.println("Tremolo Value: " + inputMessage);
    request->send(200, "text/plain", "OK");
  });

  server.on("/fuzz-slider", HTTP_GET, [] (AsyncWebServerRequest *request) {
    String inputMessage;
    // GET input1 value on <ESP_IP>/slider?value=<inputMessage>
    if (request->hasParam(PARAM_INPUT)) {
      inputMessage = request->getParam(PARAM_INPUT)->value();
      fuzzValue = inputMessage;
    }
    else {
      inputMessage = "No message sent";
    }
    Serial.println("Fuzz Value: " + inputMessage);
    request->send(200, "text/plain", "OK");
  });

  server.on("/eightBit-slider", HTTP_GET, [] (AsyncWebServerRequest *request) {
    String inputMessage;
    // GET input1 value on <ESP_IP>/slider?value=<inputMessage>
    if (request->hasParam(PARAM_INPUT)) {
      inputMessage = request->getParam(PARAM_INPUT)->value();
      eightBitValue = inputMessage;
    }
    else {
      inputMessage = "No message sent";
    }
    Serial.println("8-Bit Value: " + inputMessage);
    request->send(200, "text/plain", "OK");
  });

  // Start server
  server.begin();
  
} // ends web_interface_setup()

int disValInt = distortionValue.toInt();

// Code for setting up the codec, called in setup()
void codec_setup()
{
  // General setup needed
  //int* reg_address = WM8960_REG_AUDIO_INTERFACE_2
  //*reg_address = 0b0010000;
  codec.enableVREF();
  codec.enableVMID();

  // Setup signal flow to the ADC

  codec.enableLMIC();
  codec.enableRMIC();

  // Connect from INPUT1 to "n" (aka inverting) inputs of PGAs.
  codec.connectLMN1();
  codec.connectRMN1();

  // Disable mutes on PGA inputs (aka INTPUT1)
  codec.disableLINMUTE();
  codec.disableRINMUTE();

  // Set pga volumes
  //codec.setLINVOLDB(0); // Valid options are -17.25dB to +30dB (0.75dB steps)
  //codec.setRINVOLDB(0); // Valid options are -17.25dB to +30dB (0.75dB steps)
  codec.setLINVOLDB(3*disValInt);
  codec.setRINVOLDB(3*disValInt);

  // Set input boosts to get inputs 1 to the boost mixers
  codec.setLMICBOOST(WM8960_MIC_BOOST_GAIN_0DB);
  codec.setRMICBOOST(WM8960_MIC_BOOST_GAIN_0DB);

  // Connect from MIC inputs (aka pga output) to boost mixers
  codec.connectLMIC2B();
  codec.connectRMIC2B();

  // Enable boost mixers
  codec.enableAINL();
  codec.enableAINR();

  // Disconnect LB2LO (booster to output mixer (analog bypass)
  // For this example, we are going to pass audio throught the ADC and DAC
  codec.disableLB2LO();
  codec.disableRB2RO();

  // Connect from DAC outputs to output mixer
  codec.enableLD2LO();
  codec.enableRD2RO();

  // Set gainstage between booster mixer and output mixer
  // For this loopback example, we are going to keep these as low as they go
  codec.setLB2LOVOL(WM8960_OUTPUT_MIXER_GAIN_NEG_21DB); 
  codec.setRB2ROVOL(WM8960_OUTPUT_MIXER_GAIN_NEG_21DB);

  // Enable output mixers
  codec.enableLOMIX();
  codec.enableROMIX();

  // CLOCK STUFF, These settings will get you 44.1KHz sample rate, and class-d 
  // freq at 705.6kHz
  codec.enablePLL(); // Needed for class-d amp clock
  codec.setPLLPRESCALE(WM8960_PLLPRESCALE_DIV_2);
  codec.setSMD(WM8960_PLL_MODE_FRACTIONAL);
  codec.setCLKSEL(WM8960_CLKSEL_PLL);
  codec.setSYSCLKDIV(WM8960_SYSCLK_DIV_BY_2);
  codec.setBCLKDIV(4);
  codec.setDCLKDIV(WM8960_DCLKDIV_16);
  codec.setPLLN(7);
  codec.setPLLK(0x86, 0xC2, 0x26); // PLLK=86C226h
  //codec.setADCDIV(0); // Default is 000 (what we need for 44.1KHz)
  //codec.setDACDIV(0); // Default is 000 (what we need for 44.1KHz)
  codec.setWL(WM8960_WL_16BIT);
  //codec.writeRegister(0x09, 0b00100000);

  codec.enablePeripheralMode();
  //codec.enableMasterMode();
  //codec.setALRCGPIO(); // Note, should not be changed while ADC is enabled.

  // Enable ADCs and DACs
  codec.enableAdcLeft();
  codec.enableAdcRight();
  codec.enableDacLeft();
  codec.enableDacRight();
  codec.disableDacMute();

  //codec.enableLoopBack(); // Loopback sends ADC data directly into DAC
  codec.disableLoopBack();

  // Default is "soft mute" on, so we must disable mute to make channels active
  codec.disableDacMute(); 

  codec.enableHeadphones();
  codec.enableHeadphoneZeroCross();
  codec.enableOUT3MIX(); // Provides VMID as buffer for headphone ground

  Serial.println("Volume set to +0dB");
  codec.setHeadphoneVolumeDB(0.00);

  Serial.println("Codec Setup complete. Listen to left/right INPUT1 on Headphone outputs.");
}

// Code for setting up the i2s communication, called in setup()
void i2s_install() {
  // Set up I2S Processor configuration
  const i2s_driver_config_t i2s_config = {
    .mode = i2s_mode_t(I2S_MODE_MASTER | I2S_MODE_RX | I2S_MODE_TX),
    .sample_rate = 44100,
    .bits_per_sample = i2s_bits_per_sample_t(16),
    .channel_format = I2S_CHANNEL_FMT_RIGHT_LEFT,
    .communication_format = i2s_comm_format_t(I2S_COMM_FORMAT_STAND_MSB),
    .intr_alloc_flags = 0,
    .dma_buf_count = 8,
    .dma_buf_len = bufferLen,
    .use_apll = false,
    .tx_desc_auto_clear = false,
    .fixed_mclk = 0,
    .mclk_multiple = i2s_mclk_multiple_t(I2S_MCLK_MULTIPLE_DEFAULT),
    .bits_per_chan = i2s_bits_per_chan_t(I2S_BITS_PER_CHAN_DEFAULT)
  };

  i2s_driver_install(I2S_PORT, &i2s_config, 0, NULL);
}

void i2s_setpin() {
  // Set I2S pin configuration
  const i2s_pin_config_t pin_config = {
    .mck_io_num = -1,
    .bck_io_num = I2S_SCK,
    .ws_io_num = I2S_WS,
    .data_out_num = I2S_SDO,
    .data_in_num = I2S_SD
  };

  i2s_set_pin(I2S_PORT, &pin_config);
}

void setup()
{
  Serial.begin(115200);

  Wire.begin();

  if (codec.begin() == false) //Begin communication over I2C
  {
    Serial.println("The device did not respond. Please check wiring.");
    while (1); // Freeze
  }
  Serial.println("Device is connected properly.");

  codec_setup();

  // Set up I2S
  i2s_install();
  i2s_setpin();
  i2s_start(I2S_PORT);

  web_interface_setup();
}
 
float waveIt = 110;
float tWave = 125;
float direction = -0.5;
int iClip = 0;
void triangolo(){
    // Update the value based on the current direction
    if (iClip == 0){
      tWave += direction*tremoloValue.toInt();
    }
    
      // Check if the value has reached the lower or upper limit
      if (tWave <= 107) {
        iClip++;
        if(iClip == 25){
            direction = 0.5;
            iClip = 0;
            tWave = 107;
        }
      }

      else if (tWave >= 125) {
        iClip++;
        if(iClip == 25){
            direction = -0.5;
            iClip = 0;
            tWave = 125;
        }
      }
    codec.setHeadphoneVolume(tWave);
}
void chorus(float* audioData, int numBytes, int gainFactor)
{
  const float mixGain = 2*gainFactor; // pitch shift factor
  // Apply the chorus effect to each sample in the input audio data
  int numSamples = numBytes / sizeof(float);

  for (int i = 0; i < numSamples; i++) {
    // Apply pitch shift to the sample
    float ampedSample = audioData[i] * mixGain;
    // Mix the original and shifted samples
    audioData[i] =  (audioData[i] + ampedSample) + audioData[i];
  }
}

void eightbit(float* audioData, int numBytes){
  size_t bytesIn = 0;
  size_t bytesOut = 0;
  const float eightbitfactor = 10;

  int numSamples = numBytes / sizeof(float);

  for (int i = 0; i < numSamples; i++) {

    float ampedsample = audioData[i]*eightbitfactor;
    audioData[i] = audioData[2*i];
  }
}

void loop()
{
  if(eightBitValue.toInt() == 0){
  size_t bytesIn = 0;
  size_t bytesOut = 0;
  esp_err_t result = i2s_read(I2S_PORT, &sBuffer, bufferLen, &bytesIn, portMAX_DELAY);
  
  // codec.writeRegister(0x09, 0b00100000); //uncomment to apply 8-bit effect

  // If no distortion, no fuzz, and no tremolo (clean)
  if (distortionValue.toInt()== 0 && tremoloValue.toInt() == 0 && fuzzValue.toInt() == 0)
  {
    // Set PGA values to 0 dB for no distortion
    codec.setLINVOLDB(0);
    codec.setRINVOLDB(0);

    // Immediately play back what was read in
    esp_err_t result_w = i2s_write(I2S_PORT, &sBuffer, bytesIn, &bytesOut, portMAX_DELAY);

    // If there is an I2S write error, let us know on the serial terminal
    if(result_w != ESP_OK){
      Serial.print("I2S write error.");
    }
  } // ends the if-statement for no distortion, no fuzz, and no tremolo

  // If distortion and no delay and no fuzz
  if (distortionValue.toInt() != 0 && tremoloValue.toInt() == 0 && fuzzValue.toInt() == 0)
  {
      // Sets PGA values for distortion using function
      codec.setLINVOLDB(3*distortionValue.toInt());
      codec.setRINVOLDB(3*distortionValue.toInt());

     // Play back what was read in with distortion
     esp_err_t result_w = i2s_write(I2S_PORT, &sBuffer, bufferLen, &bytesOut, portMAX_DELAY);

  } // ends distortion and no delay case

  // If tremolo and no distortion and no fuzz
  if (distortionValue.toInt() == 0 && tremoloValue.toInt() != 0 && fuzzValue.toInt() == 0){

    // Set PGA values to 0 dB for no distortion
    codec.setLINVOLDB(0);
    codec.setRINVOLDB(0);
  
    triangolo();
    esp_err_t result_w = i2s_write(I2S_PORT, &sBuffer, bytesIn, &bytesOut, portMAX_DELAY);

   }

   // If fuzz and no distortion and no tremolo
   if (distortionValue.toInt() == 0 && tremoloValue.toInt() == 0 && fuzzValue.toInt() != 0)
   {
    // Set PGA values to 0 dB for no distortion
    codec.setLINVOLDB(0);
    codec.setRINVOLDB(0);

    float* audioData = (float*) &sBuffer; // Cast buffer as float*
    int numBytes = bytesIn; // Get the number of bytes of audio data
  
    chorus(audioData, numBytes, fuzzValue.toInt());
    esp_err_t result_w = i2s_write(I2S_PORT, &sBuffer, bytesIn, &bytesOut, portMAX_DELAY);

   }

  // If distortion and tremolo and no fuzz
  if (distortionValue.toInt() != 0 && tremoloValue.toInt() != 0 && fuzzValue.toInt() == 0)
  {
    //Set PGA values for distortion
    codec.setLINVOLDB(3*distortionValue.toInt());
    codec.setRINVOLDB(3*distortionValue.toInt());

    triangolo();
    esp_err_t result_w = i2s_write(I2S_PORT, &sBuffer, bufferLen, &bytesOut, portMAX_DELAY);
  }

  // If tremolo and fuzz and no distortion
  if (distortionValue.toInt() == 0 && tremoloValue.toInt() != 0 && fuzzValue.toInt() != 0)
  {
    float* audioData = (float*) &sBuffer; // Cast buffer as float*
    int numBytes = bytesIn; // Get the number of bytes of audio data
    chorus(audioData, numBytes, fuzzValue.toInt());

    triangolo();
    esp_err_t result_w = i2s_write(I2S_PORT, &sBuffer, bufferLen, &bytesOut, portMAX_DELAY);
  }
  }
  // 8-bit switch
  if (eightBitValue.toInt() != 0){
    size_t bytesIn = 0;
    size_t bytesOut = 0;
    
    esp_err_t result = i2s_read(I2S_PORT, &sBuffer, bufferLen/2, &bytesIn, portMAX_DELAY);
     // Set PGA values to 0 dB for no distortion
    codec.setLINVOLDB(0);
    codec.setRINVOLDB(0);

    float* audioData = (float*)&sBuffer; // Cast buffer as float*
    int numBytes = bytesIn; // Get the number of bytes of audio data
  
    eightbit(audioData, numBytes);

    esp_err_t result_w = i2s_write(I2S_PORT, &sBuffer, bufferLen/2, &bytesOut, portMAX_DELAY);

  }

}
